from pathlib import Path

import typer

from kurra.cli.console import console
from kurra.labels import find_missing_labels, get_missing_labels

app = typer.Typer(help="Labelling commands")
from rich.table import Table


@app.command(name="find", help="Find IRIs missing labels")
def find_command(
    f: str = typer.Argument(
        ..., help="The RDF file to fine IRIs missing labels within"
    ),
    local_context: str = typer.Option(
        None,
        "--local-context",
        "-l",
        help="An RDF file or directory of RDF files containing labels",
    ),
) -> None:
    if Path(f).is_file() or Path(f).is_dir():
        if local_context is not None:
            if Path(local_context).is_file() or Path(local_context).is_dir():
                for iri in find_missing_labels(Path(f), Path(local_context)):
                    console.print(iri)
            else:
                console.print(
                    "[red]If a value for local context is given, it must be an existing file or directory[/red]"
                )
        else:
            for iri in find_missing_labels(Path(f)):
                console.print(iri)
    else:
        console.print(
            "[red]The value for f must be an existing file or directory[/red]"
        )


@app.command(
    name="get",
    help="Gets labels for IRIs missing them from a given context or the KurrawongAI Semantic Background",
)
def get_command(
    f: str = typer.Argument(
        ..., help="The RDF file to fine IRIs missing labels within"
    ),
    local_context: str = typer.Option(
        None,
        "--local-context",
        "-l",
        help="An RDF file or directory of RDF files containing labels",
    ),
    additional_context: str = typer.Option(
        "https://fuseki.dev.kurrawong.ai/semback/sparql",
        "--additional-context",
        "-c",
        help="An additional source of labels",
    ),
    return_type: str = typer.Option(
        "graph",
        "--return-type",
        "-r",
        help="Return RDF in the longturtle format or a printed table of IRIs and labels. Either 'graph' or 'table'",
    ),
) -> None:
    iris = []
    if Path(f).is_file() or Path(f).is_dir():
        if local_context is not None:
            if Path(local_context).is_file() or Path(local_context).is_dir():
                iris = find_missing_labels(Path(f), Path(local_context))
            else:
                console.print(
                    "[red]If a value for local context is given, it must be an existing file or directory[/red]"
                )
        else:
            iris = find_missing_labels(Path(f))
    else:
        console.print(
            "[red]The value for f must be an existing file or directory[/red]"
        )

    if iris:
        if Path(additional_context).is_file() or Path(additional_context).is_dir():
            rdf = get_missing_labels(iris, Path(additional_context, return_type))
        else:
            rdf = get_missing_labels(iris, additional_context, return_type)

        if return_type == "graph":
            console.print(rdf.serialize(format="longturtle"))
        else:
            t = Table(show_header=True, header_style="b")
            t.add_column("IRI", no_wrap=True)
            t.add_column("Label", no_wrap=True)
            for iri, label in rdf.items():
                t.add_row(iri, label)
            console.print(t)
    else:
        console.print("No IRIs missing labels")



from pathlib import Path

import click
from rdflib.compare import isomorphic
from typer.testing import CliRunner

from kurra.cli import app
from kurra.utils import load_graph

runner = CliRunner()


def test_infer():
    INFER_DATA_DIR = Path(__file__).parent.parent.resolve() / "shacl" / "infer"

    result = runner.invoke(
        app,
        [
            "shacl",
            "infer",
            f"{INFER_DATA_DIR / '01-data.ttl'}",
            f"{INFER_DATA_DIR / '01-rules.srl'}",
        ],
    )

    assert "ns1:grandParent" in result.output


def test_infer_remove():
    INFER_DATA_DIR = Path(__file__).parent.parent.resolve() / "shacl" / "infer"

    result = runner.invoke(
        app,
        [
            "shacl",
            "infer",
            f"{INFER_DATA_DIR / '01-data.ttl'}",
            f"{INFER_DATA_DIR / '02-rules.srl'}",
        ],
    )

    results_graph_expected = load_graph(
        Path(__file__).parent.parent / "shacl/infer" / "02-results.ttl"
    )
    actual_output = load_graph(click.unstyle(result.output))

    assert isomorphic(results_graph_expected, actual_output)



from pathlib import Path
from textwrap import dedent

from typer.testing import CliRunner

from kurra.cli import app
from kurra.db.gsp import upload

runner = CliRunner()

LANG_TEST_VOC = Path(__file__).parent.parent.resolve() / "sparql" / "language-test.ttl"
TESTING_GRAPH = "https://example.com/testing-graph"


def test_query_db(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            GRAPH ?g {
                ?c a skos:Concept .
            } 
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            q,
        ],
    )
    assert result.exit_code == 0


def test_query_file():
    # TODO: work out why this test fails but the direct call works
    # direct call:
    # kurra sparql /Users/nick/work/kurrawong/kurra/tests/test_sparql/language-test.ttl "PREFIX skos: <http://www.w3.org/2004/02/skos/core#> SELECT * WHERE {     ?c a skos:Concept .}"
    q = dedent("""
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            ?c a skos:Concept .
        }""").replace("\n", "")

    result = runner.invoke(
        app,
        ["sparql", str(LANG_TEST_VOC), q],
    )
    assert (
        "https://example.com/demo-vocabs/language-test/lang-and-no-lang"
        in result.output
    )


def test_select(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            "SELECT * WHERE { <https://example.com/demo-vocabs/language-test> ?p ?o }",
        ],
    )
    assert "https://example.com/demo-vocabs/lan" in result.output


def test_describe(fuseki_container, http_client):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        sparql_endpoint, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    result = runner.invoke(
        app,
        [
            "sparql",
            sparql_endpoint,
            "DESCRIBE <https://example.com/demo-vocabs/language-test>",
        ],
    )
    assert "Made in Nov 2024 just for testing" in result.output


def test_fuseki_sparql_drop(fuseki_container):
    result = runner.invoke(
        app,
        [
            "db",
            "sparql",
            "DROP ALL",
            f"http://localhost:{fuseki_container.get_exposed_port(3030)}",
            "-u",
            "admin",
            "-p",
            "admin",
        ],
    )
    # assert result.exit_code == 0  # TODO: work out why this isn't returning 0
    assert result.output == ""


def test_query_file():
    LANG_TEST_VOC_PATH_STR = str(
        Path(__file__).parent.parent / "sparql/language-test.ttl"
    )

    result = runner.invoke(
        app,
        [
            "sparql",
            LANG_TEST_VOC_PATH_STR,
            str(Path(__file__).parent.parent / "sparql/q.sparql"),
        ],
    )

    assert "count" in result.output

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

        SELECT (COUNT(?c) AS ?count)
        WHERE {
            ?c a skos:Concept ;
        }
        """

    result = runner.invoke(
        app,
        [
            "sparql",
            LANG_TEST_VOC_PATH_STR,
            q,
        ],
    )

    assert "count" in result.output

    # checks to ensure a 'long file name' which is really just a textual SPARQL query is handled but fails to parse as it's not a real query
    q = "Sed ut perspiciatis unde omnis iste natus error sit voluptatem accusantium doloremque laudantium, totam rem aperiam, eaque ipsa quae ab illo inventore veritatis et quasi architecto beatae vitae dicta sunt explicabo. Nemo enim ipsam voluptatem quia voluptas sit aspernatur aut odit aut fugit, sed quia consequuntur magni dolores eos qui ratione voluptatem sequi nesciunt. Neque porro quisquam est, qui dolorem ipsum quia dolor sit amet, consectetur, adipisci velit, sed quia non numquam eius modi tempora incidunt ut labore et dolore magnam aliquam quaerat voluptatem. Ut enim ad minima veniam, quis nostrum exercitationem ullam corporis suscipit laboriosam, nisi ut aliquid ex ea commodi consequatur? Quis autem vel eum iure reprehenderit qui in ea voluptate velit esse quam nihil molestiae consequatur, vel illum qui dolorem eum fugiat quo voluptas nulla pariatur? At vero eos et accusamus et iusto odio dignissimos ducimus qui blanditiis praesentium voluptatum deleniti atque corrupti quos dolores."

    result = runner.invoke(
        app,
        [
            "sparql",
            LANG_TEST_VOC_PATH_STR,
            q,
        ],
    )

    assert "Expected one of" in str(result.exception)


def test_issue_37():
    u = """
        PREFIX ex: <http://example.com/>
        
        DELETE {
            ?x ex:description ?d
        }
        INSERT {
            ?x ex:description ?d
        }
        WHERE {
            ?x ex:description ?d
        }
        """
    result = runner.invoke(
        app,
        [
            "sparql",
            str(Path(__file__).parent / "issue-37.ttl"),
            u,
        ],
    )
    assert len(result.output.split("\n")) == 7  # same length as original issue-37.ttl



from pathlib import Path
from typing import Literal

import httpx
from rdflib import DCTERMS, RDFS, SDO, SKOS, BNode, Graph, URIRef

from kurra.sparql import query
from kurra.utils import load_graph


def find_missing_labels(
    p: Path | str | Graph, local_context: Path | Graph = None
) -> set[URIRef]:
    """Finds all the IRIs in a graph missing labels.

    If local_context is supplied - and it must be a Path to an RDF file or directory of RDF files or a Graph - then labels from that context will be used too."""

    # find all the things missing labels
    subjects_missing_labels = set()
    predicates_missing_labels = set()
    objects_missing_labels = set()
    g = load_graph(p)

    for s in g.subjects():
        if g.value(subject=s, predicate=DCTERMS.type):
            continue
        elif g.value(subject=s, predicate=RDFS.label):
            continue
        elif g.value(subject=s, predicate=SDO.name):
            continue
        elif g.value(subject=s, predicate=SKOS.prefLabel):
            continue

        if not isinstance(s, BNode):
            subjects_missing_labels.add(s)

    for s in g.predicates():
        if g.value(subject=s, predicate=DCTERMS.type):
            continue
        elif g.value(subject=s, predicate=RDFS.label):
            continue
        elif g.value(subject=s, predicate=SDO.name):
            continue
        elif g.value(subject=s, predicate=SKOS.prefLabel):
            continue

        if not isinstance(s, BNode):
            predicates_missing_labels.add(s)

    for s in g.objects():
        if g.value(subject=s, predicate=DCTERMS.type):
            continue
        elif g.value(subject=s, predicate=RDFS.label):
            continue
        elif g.value(subject=s, predicate=SDO.name):
            continue
        elif g.value(subject=s, predicate=SKOS.prefLabel):
            continue

        if isinstance(s, URIRef):
            objects_missing_labels.add(s)

    things_missing_labels = objects_missing_labels.union(
        predicates_missing_labels.union(subjects_missing_labels)
    )

    if local_context is not None:
        tx = set()

        c = load_graph(local_context)
        for t in things_missing_labels:
            if not c.value(subject=t, predicate=SDO.name):
                tx.add(t)
        return tx
    else:
        return sorted(things_missing_labels)


def get_missing_labels(
    iris: list[URIRef],
    context: Graph | str | Path = "https://fuseki.dev.kurrawong.ai/semback/sparql",
    return_type: Literal["graph", "dict"] = "graph",
) -> Graph | dict[URIRef, str]:
    """Gets labels for given IRIs from a given context"""
    values = ""
    for i in iris:
        values += f"\t<{i}>\n"
    values = "VALUES ?iri {\n" + values + "}\n\n"

    where_clause = f"""
        WHERE {{
            {values}
            
            ?iri schema:name ?label .
        }} 
        """

    if return_type == "graph":
        q = f"""
            PREFIX schema: <https://schema.org/>
            
            CONSTRUCT {{
                ?iri schema:name ?label
            }}
            {where_clause} 
            """
        return query(context, q, return_format="python")
    else:
        q = f"""
            PREFIX schema: <https://schema.org/>
            
            SELECT ?iri ?label
            {where_clause}
            """
        d = {}
        for r in query(context, q, return_format="python", return_bindings_only=True):
            d[r["iri"]] = r["label"]
        return d



from pathlib import Path

from rdflib import Graph

from kurra.labels import find_missing_labels, get_missing_labels


def test_find_missing_labels():
    lbls = find_missing_labels(Path(__file__).parent / "GeologicMaterialTypes.ttl")

    assert len(lbls) == 42

    lbls = find_missing_labels(
        Path(__file__).parent / "GeologicMaterialTypes.ttl",
        Path(__file__).parent / "labels.ttl",
    )

    assert len(lbls) == 19


def test_get_missing_labels():
    rdf = get_missing_labels(
        find_missing_labels(Path(__file__).parent / "GeologicMaterialTypes.ttl")
    )

    assert type(rdf) == Graph
    assert len(rdf) == 29

    rdf = get_missing_labels(
        find_missing_labels(Path(__file__).parent / "GeologicMaterialTypes.ttl"),
        return_type="dict",
    )

    assert type(rdf) == dict
    assert len(rdf.keys()) == 29



import json
import pickle
from enum import Enum
from pathlib import Path
from typing import Union

import httpx
from rdflib import BNode, Dataset, Graph, Literal, Namespace, URIRef
from sparqlkit import (
    QuerySubType,
    SparqlStatementType,
    SparqlType,
    UpdateSubType,
    statement_type_from_string,
)

RDF_SUFFIX_MAP = {
    ".nt": "application/n-triples",
    ".nq": "application/n-quads",
    ".owl": "application/rdf+xml",
    ".ttl": "text/turtle",
    ".trig": "application/trig",
    ".json": "application/ld+json",
    ".jsonld": "application/ld+json",
    ".xml": "application/rdf+xml",
}

RDF_FILE_SUFFIXES = {
    "turtle": ".ttl",
    "longturtle": ".ttl",
    "xml": ".rdf",
    "n-triples": ".nt",
    "json-ld": ".jsonld",
    "owl": ".owl",
}

SYSTEM_GRAPH_IRI = URIRef("https://olis.dev/SystemGraph")

OLIS = Namespace("https://olis.dev/")


class RenderFormat(str, Enum):
    original = "original"
    json = "json"
    markdown = "markdown"


def guess_format_from_data(rdf: str) -> str | None:
    if rdf is not None:
        rdf = rdf.strip()
        if rdf.startswith("PREFIX") or rdf.startswith("@prefix"):
            return "text/turtle"
        elif rdf.startswith("{") or rdf.startswith("["):
            return "application/ld+json"
        elif rdf.startswith("<?xml") or rdf.startswith("<rdf"):
            return "application/rdf+xml"
        elif rdf.startswith("<http"):
            return "application/n-triples"
        else:
            return "application/n-triples"
    else:
        return None


def load_graph(graph_path_or_str: Union[Graph, Path, str], recursive=False) -> Graph:
    """
    Presents an RDFLib Graph object from a pre-existing Graph, a pickle file, an RDF file or directory of files or RDF
    data in a string
    """
    # Pre-existing Graph
    if isinstance(graph_path_or_str, Graph):
        return graph_path_or_str

    # Pickle file
    if isinstance(graph_path_or_str, Path):
        if graph_path_or_str.is_file():
            pkl_path = graph_path_or_str.with_suffix(".pkl")
            if pkl_path.is_file():
                return pickle.load(open(pkl_path, "rb"))

    # Serialized RDF file or dir of files
    if isinstance(graph_path_or_str, Path):
        if Path(graph_path_or_str).is_file():
            if str(graph_path_or_str).endswith(".trig") or str(
                graph_path_or_str
            ).endswith(".jsonld"):
                return Dataset().parse(str(graph_path_or_str))
            return Graph().parse(graph_path_or_str)
        elif Path(graph_path_or_str).is_dir():
            g = Graph()
            if recursive:
                gl = Path(graph_path_or_str).rglob("*.ttl")
            else:
                gl = Path(graph_path_or_str).glob("*.ttl")
            for f in gl:
                if f.is_file():
                    g.parse(f)
            return g

    # A remote file via HTTP
    elif isinstance(graph_path_or_str, str) and graph_path_or_str.startswith("http"):
        return Graph().parse(graph_path_or_str)

    # RDF data in a string
    else:
        return Graph().parse(
            data=graph_path_or_str,
            format=guess_format_from_data(graph_path_or_str),
        )


def render_sparql_result(
    r: dict | str | Graph, rf: RenderFormat = RenderFormat.markdown
) -> str:
    """Renders a SPARQL result in a given render format"""
    if rf == RenderFormat.original:
        return r

    elif rf == RenderFormat.json:
        if isinstance(r, dict):
            return json.dumps(r, indent=4)
        elif isinstance(r, str):
            return json.dumps(json.loads(r), indent=4)
        elif isinstance(r, Graph):
            return r.serialize(format="json-ld", indent=4)

    elif rf == RenderFormat.markdown:
        if isinstance(r, Graph):  # CONSTRUCT: RDF GRaph
            output = "```turtle\n" + r.serialize(format="longturtle") + "```\n"
        else:  # SELECT or ASK: Python dict or JSON

            def render_sparql_value(v: dict) -> str:
                # TODO: handle v["datatype"]
                if v is None:
                    return ""
                elif isinstance(v, URIRef) or isinstance(v, str):
                    return f"[{v.split('/')[-1].split('#')[-1]}]({v})"
                elif isinstance(v, Literal):
                    return v
                elif isinstance(v, BNode):
                    return f"BN: {v:>6}"
                elif v["type"] == "uri":
                    return f"[{v['value'].split('/')[-1].split('#')[-1]}]({v['value']})"
                elif v["type"] == "literal":
                    return v["value"]
                elif v["type"] == "bnode":
                    return f"BN: {v['value']:>6}"

            if isinstance(r, str):
                r = json.loads(r)

            output = ""
            header = ["", ""]
            body = []

            if r.get("head") is not None:
                # SELECT
                if r["head"].get("vars") is not None:
                    for col in r["head"]["vars"]:
                        header[0] += f"{col} | "
                        header[1] += f"--- | "
                    output = (
                        "| " + header[0].strip() + "\n| " + header[1].strip() + "\n"
                    )

            if r.get("results"):
                if r["results"].get("bindings"):
                    for row in r["results"]["bindings"]:
                        row_cols = []
                        for k in r["head"]["vars"]:
                            v = row.get(k)
                            if v is not None:
                                # ignore the k
                                row_cols.append(render_sparql_value(v))
                            else:
                                row_cols.append("")
                        body.append(" | ".join(row_cols))

                output += "\n| ".join(body) + " |\n"

            if r.get("boolean") is not None:
                output = str(bool(r.get("boolean")))

        return output


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth)


def convert_sparql_json_to_python(
    j: Union[str, bytes, httpx.Response], return_bindings_only=False
) -> {}:
    if type(j) == str:
        r = json.loads(j)
    elif type(j) == bytes:
        r = json.loads(j.decode())
    elif type(j) == httpx.Response:
        r = j.json()

    if r.get("results") is not None:  # SELECT
        for row in r["results"]["bindings"]:
            for k, v in row.items():
                if v["type"] == "literal":
                    if v.get("datatype") is not None:
                        row[k] = Literal(v["value"], datatype=v["datatype"]).toPython()
                    else:
                        row[k] = Literal(v["value"]).toPython()
                elif v["type"] == "uri":
                    row[k] = v["value"]
        if return_bindings_only:
            r = r["results"]["bindings"]
        return r
    elif r.get("boolean") is not None:  # ASK
        if return_bindings_only:
            return bool(r["boolean"])
        else:
            return r
    else:
        return r


def sparql_statement_return_type(
    query: str, statement: SparqlStatementType | None = None
) -> str:
    statement = _ensure_statement_type(query, statement)
    if is_construct_or_describe_query(query, statement):
        return "text/turtle"
    return "application/sparql-results+json"


def statement_type_for_query(query: str) -> SparqlStatementType:
    return statement_type_from_string(query)


def _ensure_statement_type(
    query: str, statement: SparqlStatementType | None = None
) -> SparqlStatementType:
    return statement if statement is not None else statement_type_for_query(query)


def is_construct_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY
        and statement.subtype == QuerySubType.CONSTRUCT
    )


def is_describe_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY
        and statement.subtype == QuerySubType.DESCRIBE
    )


def is_select_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.QUERY and statement.subtype == QuerySubType.SELECT
    )


def is_ask_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype == QuerySubType.ASK


def is_construct_or_describe_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype in {
        QuerySubType.CONSTRUCT,
        QuerySubType.DESCRIBE,
    }


def is_select_or_ask_query(
    query: str, statement: SparqlStatementType | None = None
) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.QUERY and statement.subtype in {
        QuerySubType.SELECT,
        QuerySubType.ASK,
    }


def is_update_query(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return statement.type == SparqlType.UPDATE


def is_drop_update(query: str, statement: SparqlStatementType | None = None) -> bool:
    statement = _ensure_statement_type(query, statement)
    return (
        statement.type == SparqlType.UPDATE and statement.subtype == UpdateSubType.DROP
    )


def make_sparql_dataframe(sparql_result: dict):
    try:
        from pandas import DataFrame
    except ImportError:
        raise ValueError(
            'You selected the output format "dataframe" but the pandas Python package is not installed.'
        )

    if sparql_result.get("results") is not None:  # SELECT
        df = DataFrame(columns=sparql_result["head"]["vars"])
        for i, row in enumerate(sparql_result["results"]["bindings"]):
            new_row = {}
            for k, v in row.items():
                if v["type"] == "literal":
                    if v.get("datatype") is not None:
                        new_row[k] = Literal(
                            v["value"], datatype=v["datatype"]
                        ).toPython()
                    else:
                        new_row[k] = Literal(v["value"]).toPython()
                else:
                    new_row[k] = v["value"]
            df.loc[i] = new_row
        return df
    else:  # ASK
        df = DataFrame(columns=["boolean"])
        df.loc[0] = sparql_result["boolean"]

    return df


def add_namespaces_to_query_or_data(q: str, namespaces: dict):
    preamble = ""
    for k, v in namespaces.items():
        preamble += f"PREFIX {k}: <{v}>\n"
    preamble += "\n"
    return preamble + q


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
    timeout: int = 60,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth, timeout=timeout)


def get_system_graph(
    system_graph_source: str | Path | Dataset | Graph = None,
    http_client: httpx.Client | None = None,
):
    """Returns a System Graph, graph and can accept many source options"""
    system_graph = Graph(identifier=SYSTEM_GRAPH_IRI)
    system_graph.bind("olis", OLIS)
    if system_graph_source is None:
        # no incoming System Graph
        pass
    elif isinstance(system_graph_source, Path):
        # we have a Graph or Dataset file, so read it
        if not system_graph_source.is_file():
            raise ValueError(
                f"system_graph_source must be an existing RDF file. Value supplied was {system_graph_source}"
            )

        if system_graph_source.suffix == ".trig":
            system_graph += (
                Dataset()
                .parse(system_graph_source, format="trig")
                .get_graph(SYSTEM_GRAPH_IRI)
            )
        else:
            system_graph += load_graph(system_graph_source)
    elif isinstance(system_graph_source, Graph):
        # we have a Graph, so assume it's a System Graph and load it
        system_graph += system_graph_source
    elif isinstance(system_graph_source, Dataset):
        # we have a Dataset object, so load its system Graph
        system_graph += system_graph_source.get_graph(str(SYSTEM_GRAPH_IRI))
    elif system_graph_source and system_graph_source.startswith("http"):
        # we have a remote SPARQL Endpoint, so read the System Graph
        # this is simplified GSP get()
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = http_client.get(
            str(system_graph_source),
            params={"graph": SYSTEM_GRAPH_IRI},
            headers={"Accept": "text/turtle"},
        )

        if close_http_client:
            http_client.close()

        if r.is_success:
            system_graph += Graph().parse(data=r.text, format="turtle")
        else:
            return r.status_code
    elif system_graph_source and not system_graph_source.startswith("http"):
        system_graph += load_graph(system_graph_source)
    else:
        raise ValueError(
            "The parameter system_graph_source must be either None, a Path to an RDF Graph or Dataset serialised "
            "in Turtle or Trig, an RDFLib Graph object assumed to be a System Graph, an RDFLib Dataset object containing"
            "a System Graph or a string URL for a SPARQL Endpoint."
        )

    return system_graph


def put_system_graph(
    system_graph: Graph,
    system_graph_source: str | Path | Dataset | Graph = None,
    http_client: httpx.Client | None = None,
):
    if system_graph_source is None:
        return system_graph
    elif isinstance(system_graph_source, Path):
        if system_graph_source.suffix == ".trig":
            # TODO: deduplicate the Dataset parse in get_system_graph
            d = Dataset().parse(system_graph_source, format="trig")
            d.remove_graph(SYSTEM_GRAPH_IRI)
            d.add_graph(system_graph)
            d.serialize(destination=system_graph_source, format="trig")
        else:
            system_graph.serialize(destination=system_graph_source, format="longturtle")

        return None
    elif isinstance(system_graph_source, Graph):
        system_graph_source = system_graph
        return None
    elif isinstance(system_graph_source, Dataset):
        system_graph_source.remove_graph(SYSTEM_GRAPH_IRI)
        system_graph_source.add_graph(system_graph)
        return None
    elif system_graph_source and system_graph_source.startswith("http"):
        # this is simplified GSP put()
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = http_client.put(
            system_graph_source,
            params={"graph": SYSTEM_GRAPH_IRI},
            headers={"Content-Type": "text/turtle"},
            content=system_graph.serialize(format="text/turtle"),
        )

        if close_http_client:
            http_client.close()

        if r.is_success:
            return None
        else:
            return r.status_code
    elif system_graph_source and not system_graph_source.startswith("http"):
        return system_graph
    else:
        return None



from pathlib import Path
from unittest import result

import pytest
from rdflib import Graph
from typer.testing import CliRunner

from kurra.cli import app
from kurra.cli.commands.labels import find_command, get_command
from kurra.shacl import sync_validators

runner = CliRunner()


def test_find_command():
    LABELS_TEST_DIR = Path(__file__).parent.parent.resolve() / "labels"

    result = runner.invoke(
        app,
        ["labels", "find", f"{LABELS_TEST_DIR / 'GeologicMaterialTypes.ttl'}"],
    )

    iris = result.output.splitlines()
    assert len(iris) == 42


def test_get_command():
    LABELS_TEST_DIR = Path(__file__).parent.parent.resolve() / "labels"

    result = runner.invoke(
        app,
        ["labels", "get", f"{LABELS_TEST_DIR / 'GeologicMaterialTypes.ttl'}"],
        color=False,
    )

    assert "sdo:dateCreated" in result.output


def test_get_command_table():
    LABELS_TEST_DIR = Path(__file__).parent.parent.resolve() / "labels"

    result = runner.invoke(
        app,
        [
            "labels",
            "get",
            f"{LABELS_TEST_DIR / 'GeologicMaterialTypes.ttl'}",
            "-r",
            "table",
        ],
        color=False,
    )

    assert "Label" in result.output
    assert "http://www.w3.org/2002/07/owl#versionIRI" in result.output



from pathlib import Path

from rdflib.compare import isomorphic

from kurra.shacl import infer
from kurra.utils import load_graph


def test_rules():
    data_graph = Path(__file__).parent / "infer" / "01-data.ttl"

    rules = Path(__file__).parent / "infer" / "01-rules.srl"

    results_graph_expected = load_graph(
        Path(__file__).parent / "infer" / "01-results.ttl"
    )

    results_graph_received = infer(data_graph, rules)

    assert isomorphic(results_graph_expected, results_graph_received)


def test_rules_removed():
    data_graph = Path(__file__).parent / "infer" / "01-data.ttl"

    rules = Path(__file__).parent / "infer" / "02-rules.srl"

    results_graph_expected = load_graph(
        Path(__file__).parent / "infer" / "02-results.ttl"
    )

    results_graph_received = infer(data_graph, rules)

    assert isomorphic(results_graph_expected, results_graph_received)


def test_rules_include_base():
    data_graph = load_graph(Path(__file__).parent / "infer" / "01-data.ttl")

    rules = Path(__file__).parent / "infer" / "01-rules.srl"

    results_graph_expected = (
        load_graph(Path(__file__).parent / "infer" / "01-results.ttl") + data_graph
    )

    results_graph_received = infer(data_graph, rules, include_base=True)

    assert isomorphic(results_graph_expected, results_graph_received)



import re
from pathlib import Path

from typer.testing import CliRunner

from kurra.cli import app

runner = CliRunner()
from click.utils import strip_ansi


def strip_ansi(text):
    ansi_escape = re.compile(r"\x1b\[[0-9;]*m")
    return ansi_escape.sub("", text)


def test_fuseki_create(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    sparql_endpoint = f"http://localhost:{port}"
    dataset_name = "myds"

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "create",
            sparql_endpoint,
            dataset_name,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert f"Dataset myds created at http://localhost:{port}." in strip_ansi(
        result.output
    )


def test_fuseki_create_with_both_dataset_name_and_config_file(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    sparql_endpoint = f"http://localhost:{port}"
    dataset_name = "myds"
    config_file = Path(__file__).parent / "config.ttl"

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "create",
            sparql_endpoint,
            dataset_name,
            "--config",
            str(config_file),
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 2
    assert "Only dataset name or --config is allowed, not both." in strip_ansi(
        result.output
    )


def test_fuseki_create_with_config_file(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    sparql_endpoint = f"http://localhost:{port}"
    dataset_name = "myds"
    config_file = Path(__file__).parent / "config.ttl"

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "create",
            sparql_endpoint,
            "--config",
            str(config_file),
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert (
        f"Dataset myds created using assembler config at http://localhost:{port}."
        in strip_ansi(result.output)
    )


def test_fuseki_create_existing_dataset(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    sparql_endpoint = f"http://localhost:{port}"
    dataset_name = "ds"

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "create",
            sparql_endpoint,
            dataset_name,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 1
    assert (
        f"Failed to create dataset {dataset_name} at {sparql_endpoint}"
        in strip_ansi(result.output)
    )


def test_fuseki_delete(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    sparql_endpoint = f"http://localhost:{port}"
    dataset_name = "ds"

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "describe",
            sparql_endpoint,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert "'ds.name': '/ds'" in strip_ansi(result.output)

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "delete",
            sparql_endpoint,
            dataset_name,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert "Dataset ds deleted." in strip_ansi(result.output)

    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "describe",
            sparql_endpoint,
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert "'ds.name': '/ds'" not in strip_ansi(result.output)


def test_fuseki_describe(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    result = runner.invoke(
        app,
        [
            "db",
            "fuseki",
            "describe",
            f"http://localhost:{port}",
            "--username",
            "admin",
            "--password",
            "admin",
        ],
    )
    assert result.exit_code == 0
    assert "'ds.name': '/ds'" in strip_ansi(result.output)



import typer

from kurra.cli.console import console

app = typer.Typer(help="Olis commands")


@app.command(name="stub", help="Placeholder command")
def exists_command():
    console.print("This is the Olis API's placeholder command")



from pathlib import Path

import pytest
from typer.testing import CliRunner

from kurra.cli import app
from kurra.shacl import sync_validators

runner = CliRunner()


def shacl_valid():
    SHACL_TEST_DIR = Path(__file__).parent.parent.resolve() / "shacl"

    result = runner.invoke(
        app,
        [
            "shacl",
            "validate",
            f"{SHACL_TEST_DIR / 'vocab-valid.ttl'}",
            f"{SHACL_TEST_DIR / 'validator-vocpub-410.ttl'}",
        ],
    )

    assert result.output.strip() == "The data is valid"


def shacl_invalid():
    SHACL_TEST_DIR = Path(__file__).parent.parent.resolve() / "shacl"

    result = runner.invoke(
        app,
        [
            "shacl",
            "validate",
            f"{SHACL_TEST_DIR / 'vocab-invalid.ttl'}",
            f"{SHACL_TEST_DIR / 'validator-vocpub-410.ttl'}",
        ],
    )
    assert "The errors are:" in result.stdout


@pytest.mark.xfail
def shacl_list_validators():
    sync_validators()

    result = runner.invoke(
        app,
        [
            "shacl",
            "listv",
        ],
    )

    assert "Prez Manifest Validator" in result.output
    assert "fake-validator" not in result.output



import datetime
import json
from pathlib import Path

import httpx
import pytest
from rdflib.namespace import SKOS

from kurra.db.gsp import clear, upload
from kurra.sparql import query
from kurra.utils import RenderFormat, render_sparql_result

LANG_TEST_VOC = Path(__file__).parent / "language-test.ttl"
TESTING_GRAPH = "https://example.com/testing-graph"


def test_query_db(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    # SPARQL_ENDPOINT = "http://localhost:3030/test"
    TESTING_GRAPH = "https://example.com/testing-graph"
    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE { 
            GRAPH ?g {
                ?c a skos:Concept .
            } 
        }"""

    assert "--- | ---" in render_sparql_result(
        query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="python")
    )

    assert (
        "c"
        in json.loads(
            (
                render_sparql_result(
                    query(
                        SPARQL_ENDPOINT,
                        q,
                        http_client=http_client,
                        return_format="python",
                    ),
                    RenderFormat.json,
                )
            )
        )["head"]["vars"]
    )

    # test return format options

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT *
        WHERE {
            GRAPH ?g {
                ?c a skos:Concept ;
                    skos:prefLabel "English only"@en ;
                .
            }
        }"""
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r[0]["c"] == "https://example.com/demo-vocabs/language-test/en-only"

    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=False,
    )
    assert (
        r["results"]["bindings"][0]["c"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(SPARQL_ENDPOINT, q, http_client=http_client)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="original",
        return_bindings_only=True,
    )
    # return_bindings_only=True is ignored since return format is not "python"
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    q = "ASK {?s ?p ?o}"
    r = query(SPARQL_ENDPOINT, q, http_client=http_client)  # original, False
    assert '"boolean"' in r

    q = "ASK { GRAPH ?g { ?s ?p ?o} }"
    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="python")
    assert r["boolean"]

    q = "ASK { GRAPH ?g { ?s ?p ?o} }"
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r

    q = "ASK {?s ?p ?o}"
    # return_bindings_only=True is ignored since return format is not "python"
    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_bindings_only=True)
    assert '"boolean"' in r

    q = "ASK {?s ?p <http://nothing.com/x>}"
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert not r


def test_query_file():
    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE {
            ?c a skos:Concept ;
                skos:prefLabel ?pl ;
            .
            
            OPTIONAL {
                ?c skos:altLabel ?al .
            }
        }
        LIMIT 3"""

    assert "--- | --- | ---" in render_sparql_result(
        query(LANG_TEST_VOC, q, return_format="python")
    )

    r = query(LANG_TEST_VOC, q)
    assert (
        "pl" in json.loads(render_sparql_result(r, RenderFormat.json))["head"]["vars"]
    )

    # test return format options

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#> 
        SELECT * 
        WHERE {
            ?c a skos:Concept ;
                skos:prefLabel ?pl ;
            .

            OPTIONAL {
                ?c skos:altLabel ?al .
            }
        }
        LIMIT 3"""

    r = query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=True)
    assert r[0]["c"] == "https://example.com/demo-vocabs/language-test/en-only"

    r = query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=False)
    assert (
        r["results"]["bindings"][0]["c"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(LANG_TEST_VOC, q, return_format="original", return_bindings_only=False)
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    r = query(LANG_TEST_VOC, q, return_format="original", return_bindings_only=True)
    # ignore the return_bindings_only=True bit as return_format != "python"
    assert isinstance(r, str)
    r2 = json.loads(r)
    assert (
        r2["results"]["bindings"][0]["c"]["value"]
        == "https://example.com/demo-vocabs/language-test/en-only"
    )

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q)  # False, False
    assert r == '{"head": {}, "boolean": true}'

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q, return_format="python")
    assert r["boolean"]

    q = "ASK {?s ?p ?o}"
    r = query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=True)
    assert r

    q = "ASK {?s ?p ?o}"
    r = json.loads(query(LANG_TEST_VOC, q, return_bindings_only=True))
    # ignore the return_bindings_only=True bit as return_format != "python"
    assert r["boolean"]

    q = "ASK {?s ?p <http://nothing.com/x>}"
    r = query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=True)
    assert not r


def test_duplicates():
    rdf_data = """
    PREFIX people: <https://linked.data.gov.au/dataset/people/>
    PREFIX schema: <https://schema.org/>

    people:nick
        a
            schema:Person , 
            schema:Patient ;
        schema:name "Nick" ;
        schema:age 42 ;
        schema:parent people:george ;
    .

    people:george
        a schema:Person ; 
        schema:name "George" ;
        schema:age 70 ;    
    .
    """

    q = """
    PREFIX people: <https://linked.data.gov.au/dataset/people/>
    PREFIX schema: <https://schema.org/>

    SELECT ?p ?name
    WHERE {
        ?p 
            a schema:Person ;
            schema:name ?name ;
            schema:age ?age ;
        .

        FILTER (?age < 50)
    }
    """
    r = query(rdf_data, q, return_format="python", return_bindings_only=True)
    assert len(r) == 1


def test_auth(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    # SPARQL_ENDPOINT = "http://localhost:3030/test"

    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    x = httpx.get(
        SPARQL_ENDPOINT, params={"query": "ASK { ?s ?p ?o}"}, auth=("admin", "admin")
    )

    q = "ASK {?s ?p ?o}"
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r

    with pytest.raises(RuntimeError):
        query(
            SPARQL_ENDPOINT,
            q,
            None,
            None,
            return_format="python",
            return_bindings_only=True,
        )


def test_construct(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = """
        CONSTRUCT { ?s ?p ?o }
        WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        LIMIT 3       
        """

    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="python")
    assert len(r) == 3


def test_insert(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = """
        INSERT { 
            GRAPH <http://other> {
                ?s ?p ?o 
            }
        }
        WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        """

    r = query(SPARQL_ENDPOINT, q, http_client=http_client)
    assert r == ""

    q = """
        SELECT *
        WHERE {
            GRAPH <http://other> {
                ?cs a skos:ConceptScheme ;
            }            
        }
        """
    r = query(
        SPARQL_ENDPOINT,
        q,
        namespaces={"skos": SKOS},
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r[0]["cs"] == "https://example.com/demo-vocabs/language-test"


def test_204_response(fuseki_container, http_client):
    # DROP data from SPARQL Endpoint
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )
    r = query(
        SPARQL_ENDPOINT,
        "DROP ALL",
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r == ""

    # DROP data from a file
    with pytest.raises(NotImplementedError):
        query(
            Path(__file__).parent / "vocab.nq",
            "DROP ALL",
            return_format="python",
            return_bindings_only=True,
        )

    # DROP no data
    clear(SPARQL_ENDPOINT, "all", http_client)
    r = query(
        SPARQL_ENDPOINT,
        "DROP ALL",
        http_client=http_client,
        return_format="python",
        return_bindings_only=True,
    )
    assert r == ""


def test_describe(fuseki_container, http_client):
    # DROP data from SPARQL Endpoint
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    query(SPARQL_ENDPOINT, "DROP ALL", http_client=http_client)
    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    g = query(
        SPARQL_ENDPOINT,
        "DESCRIBE <https://example.com/demo-vocabs/language-test>",
        http_client=http_client,
        return_format="python",
    )
    assert len(g) == 16


def test_return_formats(fuseki_container, http_client):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    query(SPARQL_ENDPOINT, "DROP ALL", http_client=http_client)
    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    with pytest.raises(
        ValueError,
        match="must be either 'original', 'python' or 'dataframe'",
    ):
        g = query(
            SPARQL_ENDPOINT,
            "DESCRIBE <https://example.com/demo-vocabs/language-test>",
            http_client=http_client,
            return_format="json",
        )

    with pytest.raises(
        ValueError,
        match='Only SELECT and ASK queries can have return_format set to "dataframe"',
    ):
        g = query(
            SPARQL_ENDPOINT,
            "DESCRIBE <https://example.com/demo-vocabs/language-test>",
            http_client=http_client,
            return_format="dataframe",
        )

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        
        SELECT *
        WHERE {
            ?c
                a skos:Concept ;
                skos:prefLabel ?pl ;
            .
        
            OPTIONAL {
                ?c skos:altLabel ?al
            }
        }
        ORDER BY ?pl
        """
    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="dataframe")

    from pandas import DataFrame

    assert type(r) == DataFrame
    assert r["pl"][0] == "English prefLabel"

    r = query(Path(__file__).parent / "language-test.ttl", q, return_format="dataframe")
    assert type(r) == DataFrame
    assert isinstance(r["pl"][0], str)
    assert r["pl"][0] == "English prefLabel"

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

        SELECT (COUNT(?c) AS ?count) 
        WHERE {
            ?c a skos:Concept
        }
        ORDER BY ?c
        """
    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="dataframe")
    assert r["count"][0] == 7

    r = query(Path(__file__).parent / "language-test.ttl", q, return_format="dataframe")
    assert r["count"][0] == 7

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

        ASK 
        WHERE {
            <https://example.com/demo-vocabs/language-test> a skos:ConceptScheme .
        }
        """
    r = query(SPARQL_ENDPOINT, q, http_client=http_client, return_format="dataframe")
    assert r["boolean"][0]

    r = query(Path(__file__).parent / "language-test.ttl", q, return_format="dataframe")
    assert r["boolean"][0]


def test_deep_python_db(fuseki_container, http_client):
    """Ensures that 'deep python' is returned from DB queries

    Tests the "deep" conversion of SPARQL types to Python types at kurra.db.sparql() Line 96"""
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    query(SPARQL_ENDPOINT, "DROP ALL", http_client=http_client)
    upload(
        SPARQL_ENDPOINT, LANG_TEST_VOC, TESTING_GRAPH, False, http_client=http_client
    )

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

        SELECT (COUNT(?c) AS ?count) 
        WHERE {
            ?c 
                a skos:Concept .
        }
        ORDER BY ?pl
        """
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=False,
    )
    assert r["results"]["bindings"][0]["count"] == 7

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ?c 
                a skos:ConceptScheme ;
                schema:dateCreated ?dc ;
            . 
        }
        ORDER BY ?pl
        """
    r = query(
        SPARQL_ENDPOINT,
        q,
        http_client=http_client,
        return_format="python",
        return_bindings_only=False,
    )
    assert isinstance(r["results"]["bindings"][0]["dc"], datetime.date)


def test_deep_python_file():
    """Ensures that 'deep python' is returned from file queries"""
    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

        SELECT (COUNT(?c) AS ?count) 
        WHERE {
            ?c 
                a skos:Concept .
        }
        ORDER BY ?pl
        """
    r = query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=True)
    assert r[0]["count"] == 7

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ?c 
                a skos:ConceptScheme ;
                schema:dateCreated ?dc ;
            . 
        }
        ORDER BY ?pl
        """
    r = query(LANG_TEST_VOC, q, return_format="python")
    assert isinstance(r["results"]["bindings"][0]["dc"], datetime.date)

    d = """
        PREFIX ex: <https://exammple.org/>
        PREFIX schema: <https://schema.org/>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        
        ex:a schema:dateIssued "2025-12-11" .  # str
        
        ex:b schema:dateIssued "2025-12-11T14:30:20" .  # str
        
        ex:c schema:dateIssued "2025-12-11"^^xsd:date .  # date
        
        ex:d schema:dateIssued "2025-12-11T14:30:20"^^xsd:dateTime .  # datetime
        """

    q = """
        PREFIX ex: <https://exammple.org/>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ex:a schema:dateIssued ?di ;
            . 
        }
        """
    r = query(d, q, return_format="python")
    assert isinstance(r["results"]["bindings"][0]["di"], str)

    q = """
        PREFIX ex: <https://exammple.org/>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ex:b schema:dateIssued ?di ;
            . 
        }
        """
    r = query(d, q, return_format="python")
    assert isinstance(r["results"]["bindings"][0]["di"], str)

    q = """
        PREFIX ex: <https://exammple.org/>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ex:c schema:dateIssued ?di ;
            . 
        }
        """
    r = query(d, q, return_format="python")
    assert isinstance(r["results"]["bindings"][0]["di"], datetime.date)

    q = """
        PREFIX ex: <https://exammple.org/>
        PREFIX schema: <https://schema.org/>

        SELECT *
        WHERE {
            ex:d schema:dateIssued ?di ;
            . 
        }
        """
    r = query(d, q, return_format="python")
    assert isinstance(r["results"]["bindings"][0]["di"], datetime.datetime)


def test_query_file():
    assert (
        query(
            LANG_TEST_VOC,
            str(Path(__file__).parent / "q.sparql"),
            return_format="python",
            return_bindings_only=True,
        )[0]["count"]
        == 7
    )

    q = """
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    
        SELECT (COUNT(?c) AS ?count)
        WHERE {
            ?c a skos:Concept ;
        }
        """

    assert (
        query(LANG_TEST_VOC, q, return_format="python", return_bindings_only=True)[0][
            "count"
        ]
        == 7
    )



import datetime
import time
from pathlib import Path
from textwrap import dedent

from rdflib import Dataset, URIRef
from rdflib.namespace import RDF, SDO

import kurra.utils
from kurra.db.ogf import OLIS, SYSTEM_GRAPH_IRI, exclude, include, validate_system_graph

this_dir = Path(__file__).resolve().parent

basic_vg = dedent(
    """
    PREFIX olis: <https://olis.dev/>
    PREFIX schema: <https://schema.org/>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    
    <http://example.org/g/1>
        a olis:VirtualGraph ;
        olis:includes
            <http://background> ,
            <http://example.org/g/a> ,
            <http://example.org/g/b> ;
        schema:dateCreated "2026-01-01T17:25:32"^^xsd:dateTime ;
        schema:dateModified "2026-01-10T17:25:32"^^xsd:dateTime ;
    .    
    """
)

simple_rg_01 = dedent(
    """
    PREFIX ex: <http://example.com/>
    
    ex:a
        ex:b ex:c ;
        ex:d ex:e ;
    .
    """
)

simple_rg_02 = dedent(
    """
    PREFIX ex: <http://example.com/>

    ex:f
        ex:g ex:h ;
        ex:i ex:j ;
        ex:k ex:l ;
    .
    """
)

dataset_trig = dedent(
    """
PREFIX ex: <http://example.com/>
PREFIX olis: <https://olis.dev/>
PREFIX schema: <https://schema.org/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

<https://olis.dev/SystemGraph> {
    <http://example.org/g/1>
        a olis:VirtualGraph ;
        olis:includes
            <http://background> ,
            <http://example.org/g/a> ,
            <http://example.org/g/b> ;
        schema:dateCreated "2026-01-01T17:25:32"^^xsd:dateTime ;
        schema:dateModified "2026-01-10T17:25:32"^^xsd:dateTime ;
    .
}

<http://example.org/g/a> {
    ex:a
        ex:b ex:c ;
        ex:d ex:e ;
    .
}

<http://example.org/g/b> {
    ex:f
        ex:g ex:h ;
        ex:i ex:j ;
    .
}
    """
)


def test_include_none():
    including_graph_iri = URIRef("http://example.org/g/1")
    sg = include(
        including_graph_iri, ["http://example.org/g/x", "http://example.org/g/y"], None
    )

    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    assert (including_graph_iri, SDO.dateCreated, None) in sg


def test_include_nothing():
    remote_sg = this_dir / "ogf" / "01-add-to-nothing.ttl"
    including_graph_iri = URIRef("http://example.org/g/1")
    include(
        including_graph_iri,
        ["http://example.org/g/x", "http://example.org/g/y"],
        remote_sg,
    )

    sg = kurra.utils.load_graph(remote_sg)
    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    assert (including_graph_iri, SDO.dateCreated, None) in sg

    # reset test file
    with open(remote_sg, "w") as f:
        f.write("""PREFIX olis: <https://olis.dev/>
PREFIX schema: <https://schema.org/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>""")


def test_include_existing_vg():
    start_time = datetime.datetime.now()
    time.sleep(1)
    target_sg = this_dir / "ogf" / "02-add-to-vg.ttl"
    including_graph_iri = URIRef("http://example.org/g/1")
    include(
        including_graph_iri,
        ["http://example.org/g/x", "http://example.org/g/y"],
        target_sg,
    )

    sg = kurra.utils.load_graph(target_sg)
    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    dc = sg.value(subject=including_graph_iri, predicate=SDO.dateCreated)
    assert dc.value == datetime.datetime.strptime(
        "2026-01-01T17:25:32", "%Y-%m-%dT%H:%M:%S"
    )
    dm = sg.value(subject=including_graph_iri, predicate=SDO.dateModified)
    assert dm.value > start_time

    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/a")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/b")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/x")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/y")) in sg

    # reset test file
    with open(target_sg, "w") as f:
        f.write(basic_vg)


def test_include_convert_rg():
    target_sg = this_dir / "ogf" / "03-convert-rg.ttl"
    including_graph_iri = URIRef("http://example.org/g/1")
    new_rg_iri = URIRef(str(including_graph_iri) + "-real")
    include(
        including_graph_iri,
        ["http://example.org/g/x", "http://example.org/g/y"],
        target_sg,
    )

    sg = kurra.utils.load_graph(target_sg)
    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    assert (including_graph_iri, OLIS.includes, new_rg_iri) in sg
    assert (new_rg_iri, RDF.type, OLIS.RealGraph) in sg

    # reset test file
    with open(target_sg, "w") as f:
        f.write("""PREFIX olis: <https://olis.dev/>
PREFIX schema: <https://schema.org/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

<http://example.org/g/1>
    a olis:RealGraph ;
    schema:dateCreated "2026-01-01T17:25:32"^^xsd:dateTime ;
    schema:dateModified "2026-01-10T17:25:32"^^xsd:dateTime ;
.
    """)


def test_include_local_dataset():
    start_time = datetime.datetime.now()
    time.sleep(1)
    target_sg = this_dir / "ogf" / "04-local-dataset.trig"
    including_graph_iri = URIRef("http://example.org/g/1")
    include(
        including_graph_iri,
        ["http://example.org/g/x", "http://example.org/g/y"],
        target_sg,
    )

    local_dataset = Dataset().parse(target_sg, format="trig")
    sg = local_dataset.get_graph(SYSTEM_GRAPH_IRI)
    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    dc = sg.value(subject=including_graph_iri, predicate=SDO.dateCreated)
    assert dc.value == datetime.datetime.strptime(
        "2026-01-01T17:25:32", "%Y-%m-%dT%H:%M:%S"
    )
    dm = sg.value(subject=including_graph_iri, predicate=SDO.dateModified)
    assert dm.value > start_time

    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/a")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/b")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/x")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/y")) in sg

    # reset test file
    with open(target_sg, "w") as f:
        f.write(dataset_trig)


def test_include_sparql(fuseki_container):
    sparql_endpoint = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    including_graph_iri = URIRef("http://example.org/g/1")

    kurra.db.gsp.put(sparql_endpoint, basic_vg, str(SYSTEM_GRAPH_IRI))
    kurra.db.gsp.put(sparql_endpoint, simple_rg_01, "http://example.org/g/1")
    kurra.db.gsp.put(sparql_endpoint, simple_rg_02, "default")

    include(
        including_graph_iri,
        ["http://example.org/g/x", "http://example.org/g/y"],
        sparql_endpoint,
    )

    sg = kurra.db.gsp.get(sparql_endpoint, SYSTEM_GRAPH_IRI)
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/a")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/b")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/x")) in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/y")) in sg


def test_exclude_local_dataset():
    start_time = datetime.datetime.now()
    time.sleep(1)
    target_sg = this_dir / "ogf" / "04-local-dataset.trig"
    including_graph_iri = URIRef("http://example.org/g/1")
    exclude(
        including_graph_iri,
        ["http://example.org/g/a", "http://example.org/g/z"],
        target_sg,
    )

    local_dataset = Dataset().parse(target_sg, format="trig")
    sg = local_dataset.get_graph(SYSTEM_GRAPH_IRI)
    assert (including_graph_iri, RDF.type, OLIS.VirtualGraph) in sg
    dc = sg.value(subject=including_graph_iri, predicate=SDO.dateCreated)
    assert dc.value == datetime.datetime.strptime(
        "2026-01-01T17:25:32", "%Y-%m-%dT%H:%M:%S"
    )
    dm = sg.value(subject=including_graph_iri, predicate=SDO.dateModified)
    assert dm.value > start_time

    assert (
        including_graph_iri,
        OLIS.includes,
        URIRef("http://example.org/g/a"),
    ) not in sg
    assert (including_graph_iri, OLIS.includes, URIRef("http://example.org/g/b")) in sg

    # reset test file
    with open(target_sg, "w") as f:
        f.write(dataset_trig)


def test_validate_system_graph():
    valid_sg = dedent(
        """
        PREFIX ex: <http://example.com/>
        PREFIX olis: <https://olis.dev/>
        
        ex:rg-1
            a olis:RealGraph ;
        .
        
        ex:rg-2
            a olis:VirtualGraph ;
        .
        """
    )

    invalid_sg = dedent(
        """
        PREFIX ex: <http://example.com/>
        PREFIX olis: <https://olis.dev/>
        
        ex:rg-1
            a
                olis:RealGraph ,
                olis:VirtualGraph ;
        .
        """
    )

    v = validate_system_graph(valid_sg)
    assert v[0]

    v = validate_system_graph(invalid_sg)
    assert not v[0]



from io import TextIOBase
from pathlib import Path

import httpx
from rdflib import RDF, Graph, URIRef


class FusekiError(Exception):
    """An error that occurred while interacting with Fuseki."""

    def __init__(self, message_context: str, message: str, status_code: int) -> None:
        self.message = f"{status_code} {message_context}. {message}"
        super().__init__(self.message)


def ping(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.get(f"{server_url}/$/ping")

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to ping server at {server_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return r.text


def server(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.get(f"{server_url}/$/server")

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to get server information for server at {server_url}",
            r.text,
            r.status_code,
        )

    if close_http_client:
        http_client.close()

    return r.text


def status(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    return server(server_url, http_client=http_client)


def stats(
    server_url: str,
    name: str = None,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    url = f"{server_url}/$/stats" if name is None else f"{server_url}/$/stats/{name}"
    r = http_client.get(url)

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to get stats for server at {server_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return r.text


def backup(
    server_url: str,
    name: str,
    http_client: httpx.Client | None = None,
):
    raise NotImplementedError("backup/backups is not implemented yet")


def backups(
    server_url: str,
    name: str,
    http_client: httpx.Client | None = None,
):
    return backup(server_url, name, http_client)


def backups_list(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.get(f"{server_url}/$/backups-list")

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to get stats for server at {server_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return r.text


def sleep(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    raise NotImplementedError("sleep is not implemented yet")


def tasks(
    server_url: str,
    name: str = None,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    url = f"{server_url}/$/tasks" if name is None else f"{server_url}/$/tasks/{name}"
    r = http_client.get(url)

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to get stats for server at {server_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return r.text


def metrics(
    server_url: str,
    http_client: httpx.Client | None = None,
):
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.get(f"{server_url}/$/metrics")

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to get stats for server at {server_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return r.text


def describe(
    base_url: str,
    dataset_name: str = None,
    http_client: httpx.Client | None = None,
) -> dict:
    """
    Describe the datasetss or a single dataset in a Fuseki server instances.

    :param base_url: The base URL of the Fuseki server. E.g., http://localhost:3030
    :param dataset_name: The dataset to be described. If None (default), then all datasets will be listed
    :param http_client: The synchronous httpx client to be used. If this is not provided, a temporary one will be created.
    :raises FusekiError: If the datasets fail to list or the server responds with an invalid data structure.
    :returns: The Fuseki listing of datasets as a dictionary.
    """
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    headers = {"accept": "application/json"}
    url = (
        f"{base_url}/$/datasets/{dataset_name}"
        if dataset_name is not None
        else f"{base_url}/$/datasets"
    )
    r = http_client.get(url, headers=headers)

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to list datasets at {base_url}", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    try:
        if dataset_name is None:
            return r.json()["datasets"]
        else:
            return r.json()

    except KeyError:
        raise FusekiError(
            f"Failed to parse datasets r from {base_url}",
            r.text,
            r.status_code,
        )


def create(
    sparql_endpoint: str,
    dataset_name_or_config_file: str | TextIOBase | Path,
    dataset_type: str = "tdb2",
    http_client: httpx.Client | None = None,
) -> str:
    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    if isinstance(dataset_name_or_config_file, str):
        data = {"dbName": dataset_name_or_config_file, "dbType": dataset_type}
        r = http_client.post(f"{sparql_endpoint}/$/datasets", data=data)
        if r.status_code != 200 and r.status_code != 201:
            raise FusekiError(
                f"Failed to create dataset {dataset_name_or_config_file} at {sparql_endpoint}",
                r.text,
                r.status_code,
            )
        msg = f"{dataset_name_or_config_file} created at"
    else:
        if isinstance(dataset_name_or_config_file, TextIOBase):
            data = dataset_name_or_config_file.read()
        else:
            with open(dataset_name_or_config_file, "r") as file:
                data = file.read()

        graph = Graph().parse(data=data, format="turtle")
        fuseki_service = graph.value(
            None, RDF.type, URIRef("http://jena.apache.org/fuseki#Service")
        )
        dataset_name = graph.value(
            fuseki_service, URIRef("http://jena.apache.org/fuseki#name")
        )

        r = http_client.post(
            f"{sparql_endpoint}/$/datasets",
            content=data,
            headers={"Content-Type": "text/turtle"},
        )
        status_code = r.status_code
        if r.status_code != 200 and r.status_code != 201:
            raise FusekiError(
                f"Failed to create dataset {dataset_name} at {sparql_endpoint}",
                r.text,
                status_code,
            )

        msg = f"{dataset_name} created using assembler config at"

    if close_http_client:
        http_client.close()

    return f"Dataset {msg} {sparql_endpoint}."


def delete(
    base_url: str, dataset_name: str, http_client: httpx.Client | None = None
) -> str:
    """
    Delete a Fuseki dataset.

    :param base_url: The base URL of the Fuseki server. E.g., http://localhost:3030
    :param dataset_name: The dataset to be deleted
    :param http_client: The synchronous httpx client to be used. If this is not provided, a temporary one will be created.
    :raises FusekiError: If the dataset fails to delete.
    :returns: A message indicating the successful deletion of the dataset.
    """
    if not dataset_name:
        raise ValueError("You must supply a dataset name")

    close_http_client = False
    if http_client is None:
        http_client = httpx.Client()
        close_http_client = True

    r = http_client.delete(f"{base_url}/$/datasets/{dataset_name}")

    if r.status_code != 200:
        raise FusekiError(
            f"Failed to delete dataset '{dataset_name}'", r.text, r.status_code
        )

    if close_http_client:
        http_client.close()

    return f"Dataset {dataset_name} deleted."



from pathlib import Path
from pickle import dump, load

import httpx
from pyshacl import validate as v
from rdflib import Dataset, Graph, URIRef
from rdflib.namespace import RDF, SDO, SH
from srl.engine import RuleEngine
from srl.parser import SRLParser

import kurra.sparql
from kurra.db.gsp import get as gsp_get
from kurra.sparql import query
from kurra.utils import load_graph


def validate(
    data_file_or_dir_or_graph_or_list: Path | Graph | list[Path] | list[Graph],
    shacl_graph_or_file_or_url_or_id: Graph | Path | str | int,
    hide_warnings: bool = False,
) -> tuple[bool, Graph, str]:
    """Validates a data graph using a shapes graph.

    Args:
        data_file_or_dir_or_graph_or_list: The path to an RDF data file, a graph, a list of Paths or a list of Graphs to validate. List items will be merged
        shacl_graph_or_file_or_url_or_id: The sHACL shapes to validate with

    Returns:
        Tuple[bool, Graph, str]: The validation status, results graph and message, all from pySHACL

    Raises:
        ValueError: If the ID of the SHACL validator is invalid
        RuntimeError: If the IRI of the SHACL validator cannot be resolved locally or against the Semantic Background's validators
    """
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"

    data_graph = None
    shapes_graph = None

    def _get_shapes_from_iri(iri):
        local_validators = list_local_validators()
        for local_validator in local_validators.keys():
            if iri == local_validator:
                cv = load(open(validators_cache, "rb"))
                return cv.get_graph(URIRef(iri))

    def _get_shapes_from_id(id):
        id = int(id)
        local_validators = list_local_validators()
        max = len(local_validators.keys())
        if id < 0 or id > max:
            raise ValueError(f"shacl graph id value out of range. Must be <= {max}")
        for k, x in local_validators.items():
            if int(x["id"]) == id:
                cv = load(open(validators_cache, "rb"))
                return cv.get_graph(URIRef(k))

    # Try and resolve a validator IRI or string ID to a graph
    if isinstance(shacl_graph_or_file_or_url_or_id, str):
        if shacl_graph_or_file_or_url_or_id.startswith("http"):
            shapes_graph = _get_shapes_from_iri(shacl_graph_or_file_or_url_or_id)
        elif shacl_graph_or_file_or_url_or_id.isnumeric():
            shapes_graph = _get_shapes_from_id(shacl_graph_or_file_or_url_or_id)
        else:
            shapes_graph = get_validator_graph(shacl_graph_or_file_or_url_or_id)

    # Try and resolve an int validator ID to a graph
    elif isinstance(shacl_graph_or_file_or_url_or_id, int):
        shapes_graph = _get_shapes_from_id(shacl_graph_or_file_or_url_or_id)

    # Try and load the file/URL/path directly - Path
    else:
        shapes_graph = get_validator_graph(shacl_graph_or_file_or_url_or_id)

    # If the shapes graph is not yet loaded, try updating validators from the Semantic Background and try again
    if shapes_graph is None:
        # Try and resolve a validator IRI to a graph
        if isinstance(shacl_graph_or_file_or_url_or_id, str):
            if shacl_graph_or_file_or_url_or_id.startswith("http"):
                shapes_graph = _get_shapes_from_iri(shacl_graph_or_file_or_url_or_id)

    if shapes_graph is None:
        raise RuntimeError(
            f"Not able to load shapes graph: {shacl_graph_or_file_or_url_or_id}"
        )

    if isinstance(data_file_or_dir_or_graph_or_list, (Path, Graph)):
        data_graph = load_graph(data_file_or_dir_or_graph_or_list)
    elif isinstance(data_file_or_dir_or_graph_or_list, list):
        data_graph = Graph()
        for x in data_file_or_dir_or_graph_or_list:
            data_graph += load_graph(x)

    tf, g, msg = v(data_graph, shacl_graph=shapes_graph, allow_warnings=True)

    if hide_warnings:
        for s in g.subjects(predicate=RDF.type, object=SH.ValidationResult):
            if not g.value(subject=s, predicate=SH.resultSeverity) == SH.Violation:
                g = g - g.cbd(s)

    return tf, g, msg


def list_local_validators() -> dict[str, dict[str, int]] | None:
    """Lists SHACL validators - IRI & name - stored in the local system's calidator cache.

    This function does not connect over the Internet."""
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"

    if Path.is_file(validators_cache):
        local_validators = {}
        cv = load(open(validators_cache, "rb"))
        cv: Dataset
        validator_iris = [
            x.identifier
            for x in cv.graphs()
            if str(x.identifier) not in ["urn:x-rdflib:default"]
        ]

        validator_ids = load(open(validator_ids_cache, "rb"))

        for validator_iri in sorted(validator_iris):
            validator_id = validator_ids[validator_iri]
            validator_name = load_graph(cv.get_graph(validator_iri)).value(
                subject=validator_iri, predicate=SDO.name
            )
            local_validators[str(validator_iri)] = {
                "name": str(validator_name),
                "id": str(validator_id),
            }

        return local_validators
    else:
        return {}


def sync_validators(http_client: httpx.Client | None = None):
    """Checks the Semantic Background's read-only SPARQL Endpoint, currently https://fuseki.dev.kurrawong.ai/semback/sparql, for validators.

    It then checks local storage, using list_local_calidators(), to see which, if any, of those validators are stored locally.

    For any missing, it pulls down and stores a copy locally.
    """
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"
    semback_sparql_endpoint = "https://fuseki.dev.kurrawong.ai/semback/sparql"

    # get list of remote validators
    q = """
        PREFIX schema: <https://schema.org/>
        
        SELECT * 
        WHERE { 
          <https://data.kurrawong.ai/sb/validators> schema:hasPart ?p
        }
        """
    r = query(semback_sparql_endpoint, q, None, http_client, "python", True)

    remote_validators = [row["p"] for row in r]

    # get list of local validators
    local_validators = list_local_validators()

    # diff the lists
    unknown_validators = list(set(remote_validators) - set(local_validators.keys()))

    # prepare to cache
    if len(unknown_validators) > 0:
        if not kurra_cache.exists():
            Path(kurra_cache).mkdir()

        # get & add unknown remote validators to local
        if validators_cache.exists():
            d = load(open(validators_cache, "rb"))
        else:
            d = Dataset()

        for v in unknown_validators:
            g = gsp_get(semback_sparql_endpoint, v, http_client=http_client)
            if g == 422:
                raise NotImplementedError(
                    "The KurrawongAI Semantic Background set of validators is not available yet."
                )
            if not isinstance(g, Graph):
                raise RuntimeError(
                    f"The graph {v} was not obtained from the SPARQL Endpoint {semback_sparql_endpoint}"
                )
            d.add_graph(g)
            print(f"Caching validator {g.identifier}")

        with open(validators_cache, "wb") as f:
            dump(d, f)

        validator_ids = {}
        for i, v in enumerate(sorted([x.identifier for x in d.graphs()])):
            validator_ids[v] = i + 1

        with open(validator_ids_cache, "wb") as f2:
            print("Dumping validator IDs")
            dump(validator_ids, f2)

    local_validators = list_local_validators()

    return local_validators


def get_validator_graph(
    graph_or_file_or_url_or_id: Graph | Path | str | int,
) -> Graph | None:
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"

    # it's a local ID so look it up in cache
    if isinstance(graph_or_file_or_url_or_id, int) or (
        isinstance(graph_or_file_or_url_or_id, str)
        and graph_or_file_or_url_or_id.isdigit()
    ):
        validator_ids = load(open(validator_ids_cache, "rb"))
        validator_iris = [
            key
            for key, value in validator_ids.items()
            if value == int(graph_or_file_or_url_or_id)
        ]
        if len(validator_iris) != 1:
            raise ValueError(
                f"Could not find validator for {graph_or_file_or_url_or_id}"
            )

        cv = load(open(validators_cache, "rb"))
        cv: Dataset
        return cv.graph(URIRef(validator_iris[0]))

    # cater for CLI making paths strings
    if isinstance(graph_or_file_or_url_or_id, str):
        if Path(graph_or_file_or_url_or_id).exists():
            return load_graph(Path(graph_or_file_or_url_or_id))

    try:
        return load_graph(graph_or_file_or_url_or_id)
    except:
        return None


def check_validator_known(validator_iri: str) -> bool:
    """Checks first locally and then in the Semantic Background to if a validator, identified by IRI, is known"""
    local_validators = list_local_validators()
    for local_validator in local_validators.keys():
        if validator_iri == local_validator:
            return True

    sync_validators()

    local_validators = list_local_validators()
    for local_validator in local_validators.keys():
        if validator_iri == local_validator:
            return True

    return False


def infer(
    data: Graph | Path | str, rules: Graph | Path | str, include_base=False
) -> Graph:
    """Applies rules to the data graph and returns a graph of calculated results

    Args:
        data: the data to apply the rules to
        rules: the rules to apply, in SHACL Rules SPARQL syntax
        include_base: whether to include the data triples in output

    Returns:
    """
    data_graph = load_graph(data)

    if not isinstance(rules, (Path, str)):
        raise NotImplementedError(
            "Only SHACL Rules in files ending .srl or as a string containing the Shape Rules Language (SRL) syntax is "
            "currently supported. The RDF format will be supported soon."
        )

    if isinstance(rules, Path):
        if rules.suffix == ".srl":
            rules = rules.read_text()
        else:
            raise ValueError(
                f"You have specified an unknown file type for the rules. It must end with .srl. You supplied a file with: {rules.suffix}"
            )

    if "DELETE" in rules:
        if isinstance(rules, Path):
            rules = rules.read_text()

        return kurra.sparql.query(data_graph, rules)

    interim_result = RuleEngine(SRLParser().parse(rules)).evaluate(
        data_graph, inplace=False
    )

    if include_base:
        return interim_result
    else:
        return interim_result - data_graph



from pathlib import Path
from typing import Annotated

import typer
from rich.table import Table

import kurra.shacl
from kurra.cli.console import console
from kurra.cli.utils import (
    format_shacl_graph_as_rich_table,
)
from kurra.shacl import list_local_validators, sync_validators, validate

app = typer.Typer(help="SHACL commands")


@app.command(
    name="validate",
    help="Validate a given file or directory of RDF files using a given SHACL file or directory of files",
)
def validate_command(
    file_or_dir: Path = typer.Argument(
        ..., help="The file or directory of RDF files to be validated"
    ),
    shacl_graph_or_file_or_url_or_id: str = typer.Argument(
        ...,
        help="The file, directory of files, IRI of or the kurra ID for the SHACL graph to validate with",
    ),
    hide_warnings: Annotated[
        bool,
        typer.Option(
            "--hide-warnings", "-hw", help="Hides Shapes results of Warning and Info"
        ),
    ] = False,
) -> None:
    """Validate a given file or directory of files using a given SHACL file or directory of files"""
    valid, g, txt = validate(
        file_or_dir, shacl_graph_or_file_or_url_or_id, hide_warnings=hide_warnings
    )

    if valid:
        console.print("The data is valid")
    else:
        console.print("The data is NOT valid")
        console.print("The errors are:")
        console.print(format_shacl_graph_as_rich_table(g))


@app.command(
    name="listv",
    help="Lists all known SHACL validators",
)
def listv_command():
    l = list_local_validators()
    if l is None:
        console.print("No local validators found")
    else:
        t = Table()
        t.add_column("ID")
        t.add_column("IRI")
        t.add_column("Name")
        for k, v in list_local_validators().items():
            t.add_row(v["id"], k, v["name"])
        console.print(t)


@app.command(
    name="syncv",
    help="Synchronizes SHACL validators",
)
def syncv_command():
    sync_validators()

    console.print("Synchronizing SHACL validators")


@app.command(
    name="infer",
    help="Infer new triples from given data using SHACL Rules (SRL syntax only)",
)
def infer_command(
    data: str = typer.Argument(
        ...,
        help="The path of file to apply the rules to. Turtle files ending .ttl only",
    ),
    rules: str = typer.Argument(
        ...,
        help="The path of the file containing the rules to apply to the data. SHACL Rules ending .srl only",
    ),
    include_base: str = typer.Option(
        "false",
        "--include-base",
        "-ib",
        help="whether to include the data triples in output",
    ),
):
    data = Path(data)
    rules = Path(rules)

    if not Path(data).is_file() or not Path(data).suffix == ".ttl":
        console.print("You must provide a path to a .ttl file for the data")

    if not Path(rules).is_file() or not Path(rules).suffix == ".srl":
        console.print("You must provide a path to a .srl file for the rules")

    results_graph = kurra.shacl.infer(
        data, rules, include_base=True if include_base == "true" else False
    )
    console.print(results_graph.serialize(format="longturtle"))



import subprocess
from pathlib import Path


def test_reformat_cli():
    subprocess.check_output(
        [
            "kurra",
            "file",
            "reformat",
            "--output-format",
            "json-ld",
            str(Path(__file__).parent.parent.resolve() / "file/minimal1.ttl"),
        ]
    )

    comparison = """[
  {
    "@id": "http://example.com/a",
    "http://example.com/b": [
      {
        "@id": "http://example.com/c"
      }
    ]
  }
]"""

    output_file = Path(__file__).parent.parent.resolve() / "file/minimal1.jsonld"

    assert open(output_file).read() == comparison

    Path.unlink(output_file)
